//#define TI89_VERSION
#define HELP

#ifdef TI89_VERSION
#define USE_TI89
/*selects the correct constants for the 89 version*/
#define calc(x, y) (x)
#else
#define USE_TI92PLUS
#define USE_V200
/*selects the correct constants for the 92+ version*/
#define calc(x, y) (y)
#endif

#include <tigcclib.h>

#define RANDOM_STRENGTH 4
#define DELAY 3
#define VTILOG_DISABLED

#include "extgraph.h"
#include "extrand.h"
#include "vtilog.h"

#define VERSION_MAJOR 1
#define VERSION_MINOR 6
#define DATA_NAME SYMSTR("tessdata")

typedef struct
{
	char height : 5, width : 5, map[8][15];  /*A structure that is used for storing the boards.  The
	maximum size of a board is 8 by 15, and the actual size is given in the structure.  Each element of
	the	map only has its first three bits used, for the three primary types of tiles.*/
} BOARD;

typedef struct
{
	unsigned char x1 : 4, x2: 4, y1 : 4, y2 : 4, moved1 : 3, moved2 : 3, moved3 : 3;
	/*The structure that is used for storing the moves for undo purposes.  The fields are the coordinates
	of the beginning and end of the move and the values of the three squares that were involved in the
	move.*/
} MOVE;

typedef struct
{
	short left, moves; /*This structure stores how many tiles are left and how many moves you have
	taken.*/
} SCORE;

typedef struct
{
	unsigned char board_num : 4, difficulty : 2, tournament : 1; /*This structure stores information about
	the type of game that you are playing.*/
} GAME;

typedef struct
{
	char initials[3];
	unsigned char random_strength : 3, padding : 5;
} FILE_VERSION;

typedef struct
{
	MOVE data[256];
	short top;
} STACK;

typedef struct
{
	struct
	{
		short size;
		FILE_VERSION version;
	} header;
	struct
	{
		char random_weights[9][3][112];
		char save;
		short left;
		short moves;
		BOARD boards[9];
		STACK stack;
		GAME game_type;
		RANDOM_BUFFER buffer;
		SCORE current_score;
		SCORE scores[30][3];
	} data;
	struct
	{
		char hash[16];
		char padding;
		char type[7];
	} footer;
} DATA_FILE;

typedef struct
{
	char lcd[2][3840];
	BOARD board;
} MEMORY_BUFFER;

enum RET_VALS {RET_ESC = -1, RET_SAVE = 1}; /*an enum to help me remember what return value measn what*/

/*function prototypes*/
static char Menu(GAME *, BOARD[], SCORE[][3], char *);
static void MenuDrawBoard(BOARD);
static short Wrap(short, short, short);
static void DisplayScores(SCORE[][3], unsigned char[], unsigned char[], unsigned char[], unsigned char[],
char *);

#ifdef HELP
static void DisplayHelp(char *);
#endif

static void DrawPic(unsigned char, unsigned char, char, char, char, char);
static void Pause(void);
static char PlayGame(BOARD *, BOARD *, GAME *, char[][3][112], char *, short *, short *, STACK *, char *);
static void Shuffle(char[], short);
static void DoMove(BOARD *, BOARD *, short, short, short, short, short *);
static short FindLegalMoves(BOARD, char[][3], short, short);
static char UndoMove(STACK *, BOARD *, BOARD *, short, short, short *, short*);
static void UpdateLegalMoves(BOARD, BOARD *);
static char NumTiles(char);
static char MoveLeft(BOARD);
static char GameEndMenu(GAME, char *);
static void CheckHighScores(GAME, SCORE[][3], SCORE);

volatile char Time = 0;

DEFINE_INT_HANDLER(Int5Timer)
{
	if((*(char *)&Time) > 0)
	{
		(*(char *)&Time)--;
	}
}

void _main(void)
{
	#include "data.h"

	char exit_flag = 0;
	char hash[16];
	short old_font; /*used to save the old font so it can be set back the way it was*/
	DATA_FILE *filep = NULL;
	static FILE_VERSION Check = {{'P', 'J', 'R'}, RANDOM_STRENGTH, 0};
	INT_HANDLER OldInt1; /*stores the location of int 1 so it can be restored at exit*/
	INT_HANDLER OldInt5; /*stores the location of int 5 so it can be restored at exit*/
	MD5_CTX ctx;
	MEMORY_BUFFER *mem_buffer;
	SYM_ENTRY *file_vat; /*a vat style file pointer*/

	mem_buffer = malloc(sizeof(MEMORY_BUFFER)); /*allocate memory for LCD_save*/
	if(mem_buffer == NULL) /*check to see if there was enough memory free*/
	{
		/*if not...*/
		ST_helpMsg("ERROR: Memory Low");
		return;
	}

	file_vat = SymFindPtr(DATA_NAME, 0);
	if(file_vat != NULL)
	{
		if(file_vat->flags.bits.archived) /*if the data file was archived, unarchive
		it so it can be modified*/
		{
			EM_moveSymFromExtMem(DATA_NAME, HS_NULL);
		}
		filep = HeapDeref(file_vat->handle);
		MD5Init(&ctx);
		MD5Update(&ctx, (char *)&filep->data, sizeof(filep->data));
		MD5Final(hash, &ctx);
	}

	if(file_vat == NULL || memcmp(&filep->header.version, &Check, sizeof(FILE_VERSION))
	|| memcmp(hash, filep->footer.hash,	16))
	{
		randomize_long();

		file_vat = DerefSym(SymAdd(DATA_NAME));
		file_vat->handle = HeapAlloc(sizeof(DATA_FILE));
		if(file_vat == H_NULL)
		{
			free(mem_buffer);
			ST_helpMsg("ERROR: Memory Low");
			return;
		}
		filep = HeapDeref(file_vat->handle);
		memset(filep, 0, sizeof(DATA_FILE));

		filep->header.size = sizeof(DATA_FILE) - 2;
		memcpy(&filep->header.version, &Check, sizeof(FILE_VERSION));

		SaveRandState(&filep->data.buffer);
		memcpy(filep->data.boards, boards, sizeof(BOARD[9]));
		memcpy(filep->data.random_weights, random_weights, sizeof(char[9][3][112]));
		memcpy(filep->data.scores, scores, sizeof(SCORE[30][3]));
		memset(&filep->data.stack, 0, sizeof(filep->data.stack));
		filep->data.stack.top = -1;

		memcpy(filep->footer.type, &(char[]){0, 'D', 'a', 't', 'a', 0, OTH_TAG}, 7);
	}

	filep = HeapDeref(file_vat->handle);

	FastCopyScreen_R(LCD_MEM, mem_buffer->lcd[0]); /*save the screen*/
	OldInt1 = GetIntVec(AUTO_INT(1)); /*get the address of int 1 for restoring later*/
	OldInt5 = GetIntVec(AUTO_INT(5)); /*get the address of int 1 for restoring later*/
	SetIntVec(AUTO_INT(1), DUMMY_HANDLER); /*"turn off" int 1*/
	SetIntVec(AUTO_INT(5), Int5Timer); /*hook the timer onto int 5*/
	old_font = FontSetSys(F_4x6); /*set the font to 4x6*/

	LoadRandState(&filep->data.buffer);

	do
	{
		ClrScr(); /*clear the screen*/

		if(!filep->data.save) /*if there was not a saved game, go to the menu*/
		{
			exit_flag = Menu(&filep->data.game_type, filep->data.boards, filep->data.scores,
			mem_buffer->lcd[1]);
		}

		if(!exit_flag) /* if esc was not pressed and there was enough memory*/
		{
			ClrScr();
			if(filep->data.game_type.tournament) /*if you are playing a tournament game*/
			{
				for(; filep->data.game_type.board_num < 9; filep->data.game_type.board_num++) /*cycle
				through all the boards and keep a running total of the scores*/
				{
					exit_flag = PlayGame(filep->data.boards + filep->data.game_type.board_num,
					&mem_buffer->board, &filep->data.game_type, filep->data.random_weights,
					&filep->data.save, &filep->data.left, &filep->data.moves, &filep->data.stack,
					mem_buffer->lcd[1]);
					filep->data.current_score.left += filep->data.left; /*add up the running total*/
					filep->data.current_score.moves += filep->data.moves;
					ClrScr();
					if(exit_flag)
					{
						break;
					}
				}
			}
			else /*if you are not playing a tournament game*/
			{
				exit_flag = PlayGame(filep->data.boards + filep->data.game_type.board_num,
				&mem_buffer->board, &filep->data.game_type, filep->data.random_weights, &filep->data.save,
				&filep->data.left, &filep->data.moves, &filep->data.stack, mem_buffer->lcd[1]);
				/*just play one board*/
			}

			switch(filep->data.save) /*take care of the possible return values*/
			{
				case RET_ESC: /*if you pressed esc*/
				filep->data.save = 0;
				exit_flag = 0;
				break;

				case RET_SAVE: /*or you pressed clear to save your game*/
				break;

				default: /*if none of these other events happened*/
				if(!filep->data.game_type.tournament)
				{
					filep->data.current_score = (SCORE){filep->data.left, filep->data.moves};
				}
				CheckHighScores(filep->data.game_type, filep->data.scores, filep->data.current_score);
				/*update the high scores*/
			}
		}
	} while(!exit_flag); /*loop until told not to*/

	SaveRandState(&filep->data.buffer);

	MD5Init(&ctx);
	MD5Update(&ctx, (char *)&filep->data, sizeof(filep->data));
	MD5Final(filep->footer.hash, &ctx);

	EM_moveSymToExtMem(DATA_NAME, HS_NULL); /*archive the data file*/
	FontSetSys(old_font); /*set the font back to what it was*/
	SetIntVec(AUTO_INT(1), OldInt1); /*set int 1 back to what it was*/
	SetIntVec(AUTO_INT(5), OldInt5); /*set int 5 back to what it was*/

	if(mem_buffer != NULL)
	{
		FastCopyScreen_R(mem_buffer->lcd[0], LCD_MEM); /*restore the screen*/
		free(mem_buffer); /*free the screen buffer*/
	}
}

char Menu(GAME *typep, BOARD boards[], SCORE scores[][3], char *lcd) /*this function gets input from the
user to determine what kind of game they would like to play, and also allows them to view their best
scores*/
{
	char buffer[14]; /*a buffer for sprintf*/
	char ok = 0;
	short i;
	short key;
	short row = 0;
	short row_setting[3] = {}; /*the settings for each row*/
	static char range[3] = {1, 2, 8}; /*the alowable range for each row*/
	static unsigned char arrow[5] = {0x20, 0x10, 0xF8, 0x10, 0x20}; /*an arrow sprite*/
	static unsigned char arrow_x[3] = {calc(66, 106), calc(59, 99), calc(61, 101)}; /*the
	x coordinates to output the arrow at*/
	static unsigned char circle[5] = {0x70, 0x88, 0x88, 0x88, 0x70}; /*a small circle
	sprite*/
	static unsigned char circle_x[5] = {0, calc(114, 194), 0, calc(53, 93), calc(132,
	212)}; /*the x coordinates to output the circle at*/
	static unsigned char circle_y[5] = {24, 24, 36, 36, 36}; /*the y coordinates to output
	the circle at*/
	static unsigned char dot[3] = {0x70, 0x70, 0x70}; /*a dot to fill in the selected
	circles*/

	sprintf(buffer, "Tesserae v%d.%d", VERSION_MAJOR, VERSION_MINOR);
	DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 0, buffer, A_NORMAL);
	DrawStr(calc(52, 92), 6, "By Peter J. Rowe", A_NORMAL);
	DrawStr(calc(49, 89), 12, "mig53@yahoo.com", A_NORMAL);
	DrawStr(calc(71, 111), 18, "Mode", A_NORMAL);
	DrawStr(5, 24, "One Mosaic", A_NORMAL);
	DrawStr(calc(119, 199), 24, "Tournament", A_NORMAL);
	DrawStr(calc(64, 104), 30, "Difficulty", A_NORMAL);
	DrawStr(5, 36, "Beginner", A_NORMAL);
	DrawStr(calc(58, 98), 36, "Intermediate", A_NORMAL);
	DrawStr(calc(137, 217), 36, "Expert", A_NORMAL);
	DrawStr(calc(67, 107), 42, "Board 1", A_NORMAL);

#ifdef HELP
	DrawStr(calc(25, 65), calc(88, 116), "\x15/\x16: change \x17/\x18: select F1: help",
	A_NORMAL);
	DrawStr(calc(22, 62), calc(94, 122), "F2: best scores 2nd: start ESC: quit",
	A_NORMAL);
#else
	DrawStr(calc(39, 79), calc(88, 116), "\x15/\x16: change \x17/\x18: select",
	A_NORMAL);
	DrawStr(calc(22, 62), calc(94, 122), "F1: best scores 2nd: start ESC: quit",
	A_NORMAL);
#endif

	for(i = 0; i < 5; i++)
	{
		Sprite8_OR_R(circle_x[i], circle_y[i], 5, circle, LCD_MEM); /*output the circles*/
	}
	Sprite8_OR_R(0, 25, 3, dot, LCD_MEM); /*output the dots for the initially selected
	stuff*/
	Sprite8_OR_R(0, 37, 3, dot, LCD_MEM);
	Sprite8_OR_R(calc(66, 106), 18, 5, arrow, LCD_MEM); /*output the arrow at its initial
	position*/
	MenuDrawBoard(boards[row_setting[2]]); /*draw the inital board setup*/

	do
	{
		key = _rowread(0xFFFE); /*get the arrow row*/

		if(key & calc(5, 160)) /*if up or down was pressed*/
		{
			Sprite8_XOR_R(arrow_x[row], 18 + 12 * row, 5, arrow, LCD_MEM); /*erase the
			arrow*/
		}
		if(key & calc(1, 32)) /*if up was pressed*/
		{
			row--; /*decrement the row*/
			while(_rowread(0xFFFE) & calc(1, 32)); /*wait until up not pressed*/
		}
		if(key & calc(4, 128)) /*if down was pressed*/
		{
			row++; /*increment the row*/
			while(_rowread(0xFFFE) & calc(4, 128)); /*wait until down not pressed*/
		}
		if(key & calc(5, 160)) /*if up or down pressed*/
		{
			row = Wrap(row, 0, 2 - row_setting[0]); /*keep the row within the allowable
			range*/
			Sprite8_OR_R(arrow_x[row], 18 + 12 * row, 5, arrow, LCD_MEM); /*put the arrow at
			its new location*/
		}
		if(key & calc(10, 80)) /*if left or right was pressed*/
		{
			if(row < 2) /*if you are on one of the rows that has circles*/
			{
				Sprite8_XOR_R(circle_x[2 * row + row_setting[row]], circle_y[2 * row +
				row_setting[row]] + 1, 3, dot, LCD_MEM); /*turn off the dot*/
			}
			ScrRectFill(&(SCR_RECT){{0, 47, calc(159, 239), calc(87, 115)}}, ScrRect,
			A_REVERSE); /*erase the board and board number*/
		}
		if(key & calc(2, 16)) /*if left pressed*/
		{
			row_setting[row]--; /*decrement the setting of the row*/
			while(_rowread(0xFFFE) & calc(2, 16)); /*wait until left not pressed*/
		}
		if(key & calc(8, 64)) /*if right pressed*/
		{
			row_setting[row]++; /*increment the setting of the row*/
			while(_rowread(0xFFFE) & calc(8, 64)); /*wait until right not pressed*/
		}
		row_setting[row] = Wrap(row_setting[row], 0, range[row]); /*keep the setting of
		the row within the allowable range*/
		if(key & calc(10, 80)) /*if left or right pressed*/
		{
			if(row < 2) /*if you are in one of the rows that has circles*/
			{
				Sprite8_OR_R(circle_x[2 * row + row_setting[row]], circle_y[2 * row +
				row_setting[row]] + 1, 3, dot, LCD_MEM); /*put the dot back at its new
				location*/
			}
			if(row_setting[0] == 0) /*if you are not playing a tournament game*/
			{
				sprintf(buffer, "   Board %d   ", row_setting[2] + 1); /*print out which
				board you are going to play on*/
				DrawStr(calc(60, 100), 42, buffer, A_REPLACE);
				MenuDrawBoard(boards[row_setting[2]]); /*draw what the board looks like*/
			}
			else /*if you are playing a tournament game*/
			{
				DrawStr(calc(63, 103), 42, "All Boards", A_REPLACE); /*indidcate that you
				will be playing through all the boards*/
			}
			if (row == 2) /*if the arrow got erased by the ScrRectFill*/
			{
				Sprite8_OR_R(arrow_x[row], 18 + 12 * row, 5, arrow, LCD_MEM); /*put it
				back*/
			}
		}
		if(key & calc(16, 1)) /*if 2nd pressed*/
		{
			ok = 1; /*tell the loop to exit*/
			while(_rowread(0xFFFE) & calc(16, 1)); /*wait until 2nd not pressed*/
		}
		if(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)) /*if esc pressed*/
		{
			ok = 2; /*tell the loop to exit*/
			while(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)); /*wait until esc not
			pressed*/
		}
		if(_rowread(calc(0xFFDF, 0xFFBF)) & calc(128, 16)) /*if F1 pressed*/
		{
#ifdef HELP
			DisplayHelp(lcd); /*display the help screen*/
		}
		if(_rowread(0xFFEF) & calc(128, 16)) /*if F2 pressed*/
		{
#endif
			DisplayScores(scores, arrow, circle, dot, circle_x, lcd); /*display the best
			scores*/
		}
	} while(!ok);

	if(row_setting[0] == 0)
	{
		typep->board_num = row_setting[2];
	}
	else
	{
		typep->board_num = 0;
	}
	typep->difficulty = row_setting[1];
	typep->tournament = row_setting[0]; /*return information
	about what type of game the user selected*/

	return(ok - 1);
}

void MenuDrawBoard(BOARD board) /*daws the condensed form of the board outlines in the
menu*/
{
	short i;
	short j;
	static char block[3] = {0xE0, 0xE0, 0xE0}; /*a 3 by 3 sloid block*/

	for(i = 0; i < board.width; i++)
	{
		for(j = 0; j < board.height; j++)
		{
			if(board.map[j][i] != -1) /*if the map entry indicates that a tile will go in
			this space*/
			{
				Sprite8_OR_R(4 * i + (calc(161, 241) - 4 * board.width) / 2, 4 * j +
				(calc(137, 165) - 4 * board.height) / 2, 3, block, LCD_MEM); /*draw the
				block centered in the screen*/
			}
		}
	}
}

short Wrap(short input, short lower, short upper) /*take the input value and put it
between the lower and upper bound with wrapping around the ends*/
{
	return(input >= lower ? (input <= upper ? input : lower) : upper);
}

void DisplayScores(SCORE scores[][3], unsigned char arrow[], unsigned char circle[],
unsigned char dot[], unsigned char circle_x[], char *lcd) /*this function displays the best scores.
the scores array is an array of your best scores, the arrow circle and dot are sprites,
and circle_x is alignment data for the graphics*/
{
	char buffer[14] = {}; /*a buffer for sprintf*/
	short i;
	short key;
	short row = 0;
	short row_setting[3] = {};
	static char circle_y[5] = {12, 12, 24, 24, 24}; /*alignment data for the circles and
	dots*/
	static unsigned char arrow_x[3] = {calc(66, 106), calc(59, 99), calc(62, 102)};
	/*alignment data for the arrow*/
	static short range[3] = {1, 2, 8}; /*the allowable range for each row*/

	FastCopyScreen_R(LCD_MEM, lcd); /*save the screen*/
	ClrScr();

	DrawStr(calc(61, 101), 0, "High Scores", A_NORMAL);
	DrawStr(calc(71, 111), 6, "Mode", A_NORMAL);
	DrawStr(5, 12, "One Mosaic", A_NORMAL);
	DrawStr(calc(119, 199), 12, "Tournament", A_NORMAL);
	DrawStr(calc(64, 104), 18, "Difficulty", A_NORMAL);
	DrawStr(5, 24, "Beginner", A_NORMAL);
	DrawStr(calc(58, 98), 24, "Intermediate", A_NORMAL);
	DrawStr(calc(137, 217), 24, "Expert", A_NORMAL);
	DrawStr(calc(67, 107), 30, "Board 1", A_NORMAL);
	DrawStr(0, 36, "Place Tiles Left Moves Taken", A_NORMAL);
	DrawStr(calc(22, 62), calc(88, 116), "Left/Right: change Up/Down: select", A_NORMAL);
	DrawStr(calc(61, 101), calc(94, 122), "ESC: return", A_NORMAL);
	for(i = 0; i < 5; i++)
	{
		Sprite8_OR_R(circle_x[i], circle_y[i], 5, circle, LCD_MEM); /*output the circles*/
		if(i < 3)
		{
			/*output the three best scores for the initial setting sof the menu*/
			sprintf(buffer, "%d", i + 1);
			DrawStr(0, 42 + 6 * i, buffer, A_NORMAL);
			sprintf(buffer, "%u", scores[row_setting[0] == 1 ? 27 + row_setting[1] :
			row_setting[2] + 9 * row_setting[1]][i].left);
			DrawStr(21, 42 + 6 * i, buffer, A_NORMAL);
			sprintf(buffer, "%u", scores[row_setting[0] == 1 ? 27 + row_setting[1] :
			row_setting[2] + 9 * row_setting[1]][i].moves);
			DrawStr(55, 42 + 6 * i, buffer, A_NORMAL);
		}
	}
	Sprite8_OR_R(0, 13, 3, dot, LCD_MEM); /*output the dot in its initial position*/
	Sprite8_OR_R(0, 25, 3, dot, LCD_MEM); /*output the other dot in its initial position*/
	Sprite8_OR_R(calc(66, 106), 6, 5, arrow, LCD_MEM); /*output the arrow in its initial
	position*/

	do
	{
		key = _rowread(0xFFFE); /*read the arrow row*/

		if(key & calc(5, 160)) /*if up or down pressed*/
		{
			Sprite8_XOR_R(arrow_x[row], 6 + 12 * row, 5, arrow, LCD_MEM); /*erase the
			arrow*/
		}
		if(key & calc(1, 32)) /*if up pressed*/
		{
			row--; /*decrement the row*/
			while(_rowread(0xFFFE) & calc(1, 32)); /*wait until up not pressed*/
		}
		if(key & calc(4, 128)) /*if down pressed*/
		{
			row++; /*increment the row*/
			while(_rowread(0xFFFE) & calc(4, 128)); /*wait until down not pressed*/
		}
		if(key & calc(5, 160)) /* if up or down pressed*/
		{
			row = Wrap(row, 0, 2 - row_setting[0]); /*keep the row within the allowable
			range*/
			Sprite8_OR_R(arrow_x[row], 6 + 12 * row, 5, arrow, LCD_MEM); /*put the arrow at
			its new position*/
		}
		if(key & calc(10, 80)) /*if left or right pressed*/
		{
			if(row < 2) /*if you are in the rows with circles*/
			{
				Sprite8_XOR_R(circle_x[2 * row + row_setting[row]], circle_y[2 * row +
				row_setting[row]] + 1, 3, dot, LCD_MEM); /*turn off the dot*/
			}
			FastFilledRect_Erase_R(LCD_MEM, 0, 41, calc(159, 239), calc(87, 115)); /*erase the current set
			of best scores*/
		}
		if(key & calc(2, 16)) /*if left pressed*/
		{
			row_setting[row]--; /*decrement the row setting*/
			while(_rowread(0xFFFE) & calc(2, 16)); /*wait until left not pressed*/
		}
		if(key & calc(8, 64)) /*if right pressed*/
		{
			row_setting[row]++; /*increment the row setting*/
			while(_rowread(0xFFFE) & calc(8, 64)); /*wait until right not pressed*/
		}
		row_setting[row] = Wrap(row_setting[row], 0, range[row]); /*keep the row setting
		within the allowable range*/
		if(key & calc(10, 80)) /*if left or right pressed*/
		{
			if(row < 2) /*if you are in one of the rows with circles*/
			{
				Sprite8_OR_R(circle_x[2 * row + row_setting[row]], circle_y[2 * row +
				row_setting[row]] + 1, 3, dot, LCD_MEM); /*put the dot back in its new
				position*/
			}
			if(row_setting[0] == 0) /* if you are not looking at the tournament best
			scores*/
			{
				sprintf(buffer, "  Board %d   ", row_setting[2] + 1); /*print out the
				board number of the scores that you are looking at*/
			}
			else /*if you are looking at the scores for tournament mode*/
			{
				sprintf(buffer, "All Boards"); /*indicate that you are doing so*/
			}
			DrawStr(calc(63, 103), 30, buffer, A_REPLACE);
			Sprite8_OR_R(arrow_x[row], 6 + 12 * row, 5, arrow, LCD_MEM); /*make sure that
			the arrow did not get erased*/
			for(i = 0; i < 3; i++)
			{
				/*print out the set of best scores that you are currently looking at*/
				sprintf(buffer, "%d         ", i + 1);
				DrawStr(0, 42 + 6 * i, buffer, A_REPLACE);
				sprintf(buffer, "%u        ", scores[row_setting[0] == 1 ? 27 +
				row_setting[1] : row_setting[2] + 9 * row_setting[1]][i].left);
				DrawStr(21, 42 + 6 * i, buffer, A_REPLACE);
				sprintf(buffer, "%u        ", scores[row_setting[0] == 1 ? 27 +
				row_setting[1] : row_setting[2] + 9 *
				row_setting[1]][i].moves);
				DrawStr(55, 42 + 6 * i, buffer, A_REPLACE);
			}
		}
	} while(!(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64))); /*loop until esc pressed*/
	while(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)); /*wait until esc not pressed*/

	FastCopyScreen_R(lcd, LCD_MEM); /*restore the screen*/
}

#ifdef HELP
void DisplayHelp(char *lcd) /*not much to comment on here.  just displays the help*/
{
	FastCopyScreen_R(LCD_MEM, lcd);
	ClrScr();

	calc(
	{
		DrawStr(0, 0, "In this game, your goal is to clear the board of", A_NORMAL);
		DrawStr(0, 6, "tiles in as few moves as possible.  The controls", A_NORMAL);
		DrawStr(0, 12, "for the game are the arrow keys for", A_NORMAL);
		DrawStr(0, 18, "movement, 2nd to start and end jumps, ESC to", A_NORMAL);
		DrawStr(0, 24, "quit, diamond to undo a move, and clear to", A_NORMAL);
		DrawStr(0, 30, "save your game and exit.  If you save your", A_NORMAL);
		DrawStr(0, 36, "game, it will go back to where you were when", A_NORMAL);
		DrawStr(0, 42, "you start the program again.  When you press", A_NORMAL);
		DrawStr(0, 48, "2nd to start a jump, the legal destination", A_NORMAL);
		DrawStr(0, 54, "squares will flash.  If you press 2nd with the", A_NORMAL);
		DrawStr(0, 60, "cursor on any of these flashing squares, the", A_NORMAL);
		DrawStr(0, 66, "jump will be completed, otherwise the jump", A_NORMAL);
		DrawStr(0, 72, "will be aborted.  The tiles in this game have", A_NORMAL);
		DrawStr(0, 78, "three types          , which will be known as the", A_NORMAL);
		DrawPic(41, 78, 1, 1, 0, 0);
		DrawPic(47, 78, 2, 1, 0, 0);
		DrawPic(53, 78, 4, 1, 0, 0);
		DrawStr(0, 84, "primaries.  If two different primaries land on", A_NORMAL);
		DrawStr(44, 94, "Press 2nd to continue", A_NORMAL);
		Pause();
		ClrScr();

		DrawStr(0, 0, "the same square, they add together to form a", A_NORMAL);
		DrawStr(0, 6, "secondary.  Examples of this are      +     =     ,", A_NORMAL);
		DrawPic(114, 6, 1, 1, 0, 0);
		DrawPic(129, 6, 2, 1, 0, 0);
		DrawPic(143, 6, 3, 1, 0, 0);
		DrawStr(0, 12, "     +     =     , and      +     =     .  If the third",
		A_NORMAL);
		DrawPic(3, 12, 1, 1, 0, 0);
		DrawPic(17, 12, 4, 1, 0, 0);
		DrawPic(31, 12, 5, 1, 0, 0);
		DrawPic(59, 12, 2, 1, 0, 0);
		DrawPic(74, 12, 4, 1, 0, 0);
		DrawPic(88, 12, 6, 1, 0, 0);
		DrawStr(0, 18, "primary lands on that square, it will add to", A_NORMAL);
		DrawStr(0, 24, "the primaries already there to form the", A_NORMAL);
		DrawStr(0, 30, "tertiary,      +     +     =     .  There are several", A_NORMAL);
		DrawPic(35, 30, 1, 1, 0, 0);
		DrawPic(50, 30, 2, 1, 0, 0);
		DrawPic(64, 30, 4, 1, 0, 0);
		DrawPic(78, 30, 7, 1, 0, 0);
		DrawStr(0, 36, "rules about which jumps the tiles can make,", A_NORMAL);
		DrawStr(0, 42, "which will be explained next.  The first rule is", A_NORMAL);
		DrawStr(0, 48, "that primaries can only jump over other", A_NORMAL);
		DrawStr(0, 54, "primaries, secondaries which contain the", A_NORMAL);
		DrawStr(0, 60, "primary, and the tertiary, and they can only", A_NORMAL);
		DrawStr(0, 66, "land on blank spaces, primaries, or", A_NORMAL);
		DrawStr(0, 72, "secondaries that do not contain the primary", A_NORMAL);
		DrawStr(0, 78, "that is doing the jumping.  When a primary", A_NORMAL);
		DrawStr(0, 84, "jumps over another primary, the primary that", A_NORMAL);
		DrawStr(44, 94, "Press 2nd to continue", A_NORMAL);
		Pause();
		ClrScr();

		DrawStr(0, 0, "is jumped over is removed.  An example of this", A_NORMAL);
		DrawStr(0, 6, "is          becoming          after a jump.  When a", A_NORMAL);
		DrawPic(7, 6, 1, 1, 0, 0);
		DrawPic(13, 6, 2, 1, 0, 0);
		DrawPic(71, 6, 1, 1, 0, 0);
		DrawStr(0, 12, "primary jumps over a secondary, the primary", A_NORMAL);
		DrawStr(0, 18, "is removed from the secondary, leaving a", A_NORMAL);
		DrawStr(0, 24, "different primary.  For example,          will", A_NORMAL);
		DrawPic(115, 24, 1, 1, 0, 0);
		DrawPic(121, 24, 5, 1, 0, 0);
		DrawStr(0, 30, "become          .  When a primary jumps over the", A_NORMAL);
		DrawPic(34, 30, 4, 1, 0, 0);
		DrawPic(40, 30, 1, 1, 0, 0);
		DrawStr(0, 36, "tertiary, the primary is removed from the", A_NORMAL);
		DrawStr(0, 42, "tertiary, leaving a secondary.  For example,", A_NORMAL);
		DrawStr(0, 48, "         will become          .  The next rule is that",
		A_NORMAL);
		DrawPic(0, 48, 1, 1, 0, 0);
		DrawPic(6, 48, 7, 1, 0, 0);
		DrawPic(68, 48, 6, 1, 0, 0);
		DrawPic(74, 48, 1, 1, 0, 0);
		DrawStr(0, 54, "secondaries can only jump over identical", A_NORMAL);
		DrawStr(0, 60, "secondaries or the tertiary.  For example,", A_NORMAL);
		DrawStr(0, 66, "         will become          or          will become", A_NORMAL);
		DrawPic(0, 66, 3, 1, 0, 0);
		DrawPic(6, 66, 3, 1, 0, 0);
		DrawPic(74, 66, 3, 1, 0, 0);
		DrawPic(90, 66, 3, 1, 0, 0);
		DrawPic(96, 66, 7, 1, 0, 0);
		DrawStr(0, 72, "         .  Secondaries can only land on a blank", A_NORMAL);
		DrawPic(6, 72, 4, 1, 0, 0);
		DrawPic(12, 72, 3, 1, 0, 0);
		DrawStr(0, 78, "space or primaries that are not contained in", A_NORMAL);
		DrawStr(0, 84, "the secondary that is jumping.  The last rule is", A_NORMAL);
		DrawStr(44, 94, "Press 2nd to continue", A_NORMAL);
		Pause();
		ClrScr();

		DrawStr(0, 0, "that tertiaries can only jump over tertiaries,", A_NORMAL);
		DrawStr(0, 6, "and can only land on blank spaces, an example", A_NORMAL);
		DrawStr(0, 12, "of which is          becoming          .  If you are", A_NORMAL);
		DrawPic(38, 12, 7, 1, 0, 0);
		DrawPic(44, 12, 7, 1, 0, 0);
		DrawPic(102, 12, 7, 1, 0, 0);
		DrawStr(0, 18, "confused by all the rules about what can jump", A_NORMAL);
		DrawStr(0, 24, "over what, just play the game and you will", A_NORMAL);
		DrawStr(0, 30, "figure them out.  You only need to know the", A_NORMAL);
		DrawStr(0, 36, "rules to help you spot legal moves and plan", A_NORMAL);
		DrawStr(0, 42, "ahead more efficiently.", A_NORMAL);
		DrawStr(44, 94, "Press 2nd to continue", A_NORMAL);
		Pause();
	}
	,
	{
		DrawStr(0, 0, "In this game, your goal is to clear the board of tiles in as few"
		" moves as", A_NORMAL);
		DrawStr(0, 6, "possible.  The controls for the game are the arrow keys for",
		A_NORMAL);
		DrawStr(0, 12, "movement, 2nd to start and end jumps, ESC to quit, diamond to"
		" undo a", A_NORMAL);
		DrawStr(0, 18, "move, and clear to save your game and exit.  If you save your"
		" game, it", A_NORMAL);
		DrawStr(0, 24, "will go back to where you were when you start the program again.",
		A_NORMAL);
		DrawStr(0, 30, "When you press 2nd to start a jump, the legal destination"
		" squares will", A_NORMAL);
		DrawStr(0, 36, "flash.  If you press 2nd with the cursor on any of these"
		" flashing", A_NORMAL);
		DrawStr(0, 42, "squares, the jump will be completed, otherwise the jump will be",
		A_NORMAL);
		DrawStr(0, 48, "aborted.  The tiles in this game have three types          ,"
		" which will be", A_NORMAL);
		DrawPic(165, 48, 1, 1, 0, 0);
		DrawPic(172, 48, 2, 1, 0, 0);
		DrawPic(179, 48, 4, 1, 0, 0);
		DrawStr(0, 54, "known as the primaries.  If two different primaries land on the"
		" same", A_NORMAL);
		DrawStr(0, 60, "square, they merge to form a secondary.  Examples of this are",
		A_NORMAL);
		DrawStr(0, 66, "     +     =     ,      +     =     , and      +     =     .  If"
		" the third primary lands", A_NORMAL);
		DrawPic(3, 66, 1, 1, 0, 0);
		DrawPic(17, 66, 2, 1, 0, 0);
		DrawPic(31, 66, 3, 1, 0, 0);
		DrawPic(45, 66, 1, 1, 0, 0);
		DrawPic(60, 66, 4, 1, 0, 0);
		DrawPic(74, 66, 5, 1, 0, 0);
		DrawPic(103, 66, 2, 1, 0, 0);
		DrawPic(117, 66, 4, 1, 0, 0);
		DrawPic(131, 66, 6, 1, 0, 0);
		DrawStr(0, 72, "on that square, it will merge with the primaries already there"
		" to", A_NORMAL);
		DrawStr(0, 78, "form the tertiary,      +     +     =     .  There are several"
		" rules about", A_NORMAL);
		DrawPic(68, 78, 1, 1, 0, 0);
		DrawPic(82, 78, 2, 1, 0, 0);
		DrawPic(96, 78, 4, 1, 0, 0);
		DrawPic(110, 78, 7, 1, 0, 0);
		DrawStr(0, 84, "which jumps the tiles can make, which will be explained next."
		"  The", A_NORMAL);
		DrawStr(0, 90, "first rule is that primaries can only jump over other primaries,",
		A_NORMAL);
		DrawStr(0, 96, "secondaries which contain the primary, and the tertiary, and"
		" they can", A_NORMAL);
		DrawStr(0, 102, "only land on blank spaces, primaries, or secondaries that do"
		" not", A_NORMAL);
		DrawStr(0, 108, "contain the primary that is doing the jumping.  When a primary"
		" jumps", A_NORMAL);
		DrawStr(0, 114, "over another primary, the primary that is jumped over is"
		" removed.", A_NORMAL);
		DrawStr(84, 122, "Press 2nd to continue", A_NORMAL);
		Pause();
		ClrScr();

		DrawStr(0, 0, "An example of this is          becoming          after a jump."
		"  When a", A_NORMAL);
		DrawPic(71, 0, 1, 1, 0, 0);
		DrawPic(77, 0, 2, 1, 0, 0);
		DrawPic(135, 0, 1, 1, 0, 0);
		DrawStr(0, 6, "primary jumps over a secondary, the primary is removed from the",
		A_NORMAL);
		DrawStr(0, 12, "secondary, leaving a different primary.  For example,"
		"          will", A_NORMAL);
		DrawPic(188, 12, 1, 1, 0, 0);
		DrawPic(194, 12, 5, 1, 0, 0);
		DrawStr(0, 18, "become          .  When a primary jumps over the tertiary, the"
		" primary", A_NORMAL);
		DrawPic(34, 18, 4, 1, 0, 0);
		DrawPic(40, 18, 1, 1, 0, 0);
		DrawStr(0, 24, "is removed from the tertiary, leaving a secondary.  For example,",
		A_NORMAL);
		DrawStr(0, 30, "         will become          .  The next rule is that"
		" secondaries can only", A_NORMAL);
		DrawPic(0, 30, 1, 1, 0, 0);
		DrawPic(6, 30, 7, 1, 0, 0);
		DrawPic(68, 30, 6, 1, 0, 0);
		DrawPic(74, 30, 1, 1, 0, 0);
		DrawStr(0, 36, "jump over identical secondaries or the tertiary.  For example,",
		A_NORMAL);
		DrawPic(217, 36, 3, 1, 0, 0);
		DrawPic(223, 36, 3, 1, 0, 0);
		DrawStr(0, 42, "will become          or          will become          ."
		"  Secondaries can only", A_NORMAL);
		DrawPic(56, 42, 3, 1, 0, 0);
		DrawPic(72, 42, 3, 1, 0, 0);
		DrawPic(78, 42, 7, 1, 0, 0);
		DrawPic(140, 42, 4, 1, 0, 0);
		DrawPic(146, 42, 3, 1, 0, 0);
		DrawStr(0, 48, "land on a blank space or primaries that are not contained in the",
		A_NORMAL);
		DrawStr(0, 54, "secondary that is jumping.  The last rule is that tertiaries can"
		" only", A_NORMAL);
		DrawStr(0, 60, "jump over tertiaries, and can only land on blank spaces, an"
		" example", A_NORMAL);
		DrawStr(0, 66, "of which is          becoming          .  If you are confused by"
		" all the rules", A_NORMAL);
		DrawPic(38, 66, 7, 1, 0, 0);
		DrawPic(44, 66, 7, 1, 0, 0);
		DrawPic(102, 66, 7, 1, 0, 0);
		DrawStr(0, 72, "about what can jump over what, just play the game and you will",
		A_NORMAL);
		DrawStr(0, 78, "figure them out.  You only need to know the rules to help you"
		" spot", A_NORMAL);
		DrawStr(0, 84, "legal moves and plan ahead more efficiently.", A_NORMAL);
		DrawStr(84, 122, "Press 2nd to continue", A_NORMAL);
		Pause();
	}
	);

	FastCopyScreen_R(lcd, LCD_MEM);
}
#endif

void DrawPic(unsigned char x, unsigned char y, char type, char mode, char width, char
height) /*this routine does most of the graphics for the game.  if mode is 0, x and y are
the board coordinates of the tile.  if mode is 1, x and y are the screen coordinates of
the tile.  the first three bits of type determine wich tiles are displayed, and width and
height are used to center the board.*/
{
	short xx = (mode == 1 ? x : calc(80, 120) - 4 * width + 8 * x + 2);
	short yy = (mode == 1 ? y : (calc(105, 133) - 8 * height) / 2 + 8 * y + 2);
	static unsigned char blank[5] = {0x07, 0x07, 0x07, 0x07, 0x07}; /*clears out the part
	of the square where the symbols are displayed*/
	static unsigned char blank2[7] = {0xFE, 0x82, 0x82, 0x82, 0x82, 0x82, 0xFE}; /*inverts
	the rest of the square*/
	static unsigned char circle[5] = {0x70, 0x88, 0x88, 0x88, 0x70}; /*the sprite for the
	circle tile*/
	static unsigned char cross[5] = {0x20, 0x20, 0xF8, 0x20, 0x20}; /*the sprite for the
	cross tile*/
	static unsigned char inverse[7] = {0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE}; /*used
	to invert the square*/
	static unsigned char square[5] = {0, 0x70, 0x50, 0x70, 0}; /*the square tile*/
	static unsigned short box[9] = {0xFF80, 0x8080, 0x8080, 0x8080, 0x8080, 0x8080,
	0x8080, 0x8080, 0xFF80}; /*the border around each square*/

	if(type < 8) /*if the function is being called to output a tile*/
	{
		Sprite8_AND_R(xx, yy, 5, blank, LCD_MEM); /*erase the place where the tile will go*/
		if(type & 1) /*if the first bit of the type is set*/
		{
			Sprite8_OR_R(xx, yy, 5, circle, LCD_MEM); /*output a circle*/
		}
		if(type & 2) /*if the second byte of the type is set*/
		{
			Sprite8_OR_R(xx, yy, 5, square, LCD_MEM); /*output a square*/
		}
		if(type & 4) /*if the third byte of the type is set*/
		{
			Sprite8_OR_R(xx, yy, 5, cross, LCD_MEM); /*output a cross*/
		}
	}
	else if(type == 8) /*if the function is being called to output the box around each
	square*/
	{
		Sprite16_OR_R(xx - 2, yy - 2, 9, box, LCD_MEM); /*output a box*/
	}
	else if(type == 9) /*if the function is being called to display the cursor*/
	{
		Sprite8_XOR_R(xx - 1, yy - 1, 7, inverse, LCD_MEM); /*invert the square*/
	}
	else
	{
		Sprite8_XOR_R(xx - 1, yy - 1, 7, blank2, LCD_MEM); /*this cleans stuff up if a tile
		is output to a square that is currently inverted*/
	}
}

void Pause(void) /*waits for you to press second*/
{
	while(_rowread(0xFFFE) & calc(16, 1)); /*wait until second not pressed*/
	while(!(_rowread(0xFFFE) & calc(16, 1))); /*wait until second pressed*/
	while(_rowread(0xFFFE) & calc(16, 1)); /*wait until second not pressed*/
}

char PlayGame(BOARD *boardp, BOARD *moves_leftp, GAME *typep, char random_weights[][3][112], char *savep,
short *leftp, short *movesp, STACK *stack, char *lcd) /*this function does the actual gameplay
part.*/
{
	char blink = 0; /*indicates wether or not the blink routine is active*/
	char blinker = 0; /*controls the blink rate*/
	char buffer[18]; /*a buffer for sprintf*/
	char legal_moves[3][3]; /*an array used to communicate where the legal moves are*/
	char secondary[3] = {3, 5, 6}; /*the set of secondary tiles*/
	short i = 0;
	short j;
	short key;
	short go_on = 0;
	short x;
	short y;
	static char random_length[9] = {48, 60, 72, 84, 64, 80, 112, 84, 61}; /*indicates how
	many entries there are in the random weights table for each board*/

	moves_leftp->height = boardp->height; /*set the height of the moves left board to be
	the same as the tile board*/
	moves_leftp->width = boardp->width; /*set the width of the moves left board to be the
	same as the tile board*/
	*leftp = 0;
	*movesp = 0;

	if(*savep)
	{
		*savep = 0;
	}
	else
	{
		Shuffle(random_weights[typep->board_num][typep->difficulty],
		random_length[typep->board_num]); /*shuffle the random weight list for the current
		board*/

		for(y = 0; y < boardp->height; y++)
		{
			for(x = 0; x < boardp->width; x++)
			{
				if(boardp->map[y][x] != -1) /*if the current square is within the shape of
				the board*/
				{
					switch(random_weights[typep->board_num][typep->difficulty][i])
					{
						case 0: /*put a primary in the current square*/
						boardp->map[y][x] = 1 << random_long(3);
						(*leftp)++;
						break;

						case 1: /*put a secondary in the current square*/
						boardp->map[y][x] = secondary[random_long(3)];
						(*leftp) += 2;
						break;

						case 2: /*put the tertiary in the current square*/
						boardp->map[y][x] = 7;
						(*leftp) += 3;
					}
					i++; /*switch to the next random weight entry*/
				}
			}
		}
	}

	for(y = 0; y < boardp->height; y++)
	{
		for(x = 0; x < boardp->width; x++)
		{
			moves_leftp->map[y][x] = FindLegalMoves(*boardp, legal_moves, x, y); /*fill up
			the legal moves board*/
			if(boardp->map[y][x] != -1) /*if the space is within the shape of the board*/
			{
				DrawPic(x, y, 8, 0, boardp->width, boardp->height); /*draw the squares
				around each space*/
				DrawPic(x, y, boardp->map[y][x], 0, boardp->width, boardp->height); /*draw
				the appropriate tile for the space*/
			}
		}
	}
	x = y = 0; /*start out in the upper left corner*/
	while(boardp->map[y][x] == -1) /*if the upper left corner is outside the shape of the
	board, find the first square to the right that is in the shape*/
	{
		x++;
	}
	DrawPic(x, y, 9, 0, boardp->width, boardp->height); /*draw the cursor at tis initial
	position*/
	sprintf(buffer, "Board %d", typep->board_num + 1); /*print out which board you are
	playing on*/
	DrawStr(0, 0, buffer, A_NORMAL);
	sprintf(buffer, "Tiles Left: %d", *leftp); /*print out how many tiles are left*/
	DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 0, buffer, A_NORMAL);
	sprintf(buffer, "Moves: %d", *movesp); /*print out how many moves you have taken*/
	DrawStr(calc(160, 240) - DrawStrWidth(buffer, F_4x6), 0, buffer, A_NORMAL);
	do
	{
		key = _rowread(0xFFFE); /*read the arrow keys*/

		if(key & calc(15, 240)) /*if any arrow key pressed*/
		{
			DrawPic(x, y, 9, 0, boardp->width, boardp->height); /*erase the cursor*/
		}
		if(key & calc(1, 32)) /*if up pressed*/
		{
			y--;
			if(y < 0)
			{
				y = boardp->height - 1;
			}
			if(boardp->map[y][x] == -1) /*make sure that the new coordinate is within the
			shape of the board*/
			{
				do
				{
					y--;
					if(y < 0)
					{
						y = boardp->height - 1;
					}
				} while(boardp->map[y][x] == -1);
			}
			while(_rowread(0xFFFE) & calc(1, 32)); /*wait until up not pressed*/
		}
		if(key & calc(2, 16)) /*if left pressed*/
		{
			x--;
			if(x < 0)
			{
				x = boardp->width - 1;
			}
			if(boardp->map[y][x] == -1) /*make sure that the new coordinate is within the
			shape of the board*/
			{
				do
				{
					x--;
					if(x < 0)
					{
						x = boardp->width - 1;
					}
				} while(boardp->map[y][x] == -1);
			}
			while(_rowread(0xFFFE) & calc(2, 16)); /*wait until left not pressed*/
		}
		if(key & calc(4, 128)) /*if down pressed*/
		{
			y++;
			if(y == boardp->height)
			{
				y = 0;
			}
			if(boardp->map[y][x] == -1) /*make sure that the new coordinate is within the
			shape of the board*/
			{
				do
				{
					y++;
					if(y == boardp->height)
					{
						y = 0;
					}
				} while(boardp->map[y][x] == -1);
			}
			while(_rowread(0xFFFE) & calc(4, 128)); /*wait until down not pressed*/
		}
		if(key & calc(8, 64)) /*if right pressed*/
		{
			x++;
			if(x == boardp->width)
			{
				x = 0;
			}
			if(boardp->map[y][x] == -1) /*make sure that the new coordinate is within the
			shape of the board*/
			{
				do
				{
					x++;
					if(x == boardp->width)
					{
						x = 0;
					}
				} while(boardp->map[y][x] == -1);
			}
			while(_rowread(0xFFFE) & calc(8, 64)); /*wait until right not pressed*/
		}
		if(key & calc(15, 240)) /*if an arrow key pressed*/
		{
			DrawPic(x, y, 9, 0, boardp->width, boardp->height); /*put the cursor at its
			new position*/
		}
		if(key & calc(16, 1)) /*if second pressed*/
		{
			if(blink) /*if a move is waiting to be completed*/
			{
				if(blinker) /*if the time for the squares to blink has expired*/
				{
					blinker = 0; /*reset the time blink state*/
					for(i = -1; i <= 1; i++)
					{
						for(j = -1; j <= 1; j++)
						{
							if(legal_moves[j + 1][i + 1])
							{
								DrawPic(stack->data[stack->top].x1 + 2 * i, stack->data[stack->top].y1 +
								2 * j, 9, 0, boardp->width, boardp->height); /*draw the blinking squares*/
							}
						}
					}
				}
				if(legal_moves[(y - stack->data[stack->top].y1) / 2 + 1][(x - stack->data[stack->top].x1)
				/ 2 + 1] && (x - stack->data[stack->top].x1) / 2 * 2 == x - stack->data[stack->top].x1 &&
				(y - stack->data[stack->top].y1) / 2 * 2 == y - stack->data[stack->top].y1 && (x !=
				stack->data[stack->top].x1 || y != stack->data[stack->top].y1) && abs(x -
				stack->data[stack->top].x1) <= 2 && abs(y - stack->data[stack->top].y1) <= 2) /*if second
				was pressed in the appropriate spot to complete a legal move*/
				{
					dbg_printfln("writing end to top = %d", stack->top);
					stack->data[stack->top].x2 = x;
					/*the end of the move is the current cursor position*/
					stack->data[stack->top].y2 = y;

					/*store the values of the three squares involved in the move*/
					stack->data[stack->top].moved1 =
					boardp->map[stack->data[stack->top].y1][stack->data[stack->top].x1];
					stack->data[stack->top].moved2 = boardp->map[(y + stack->data[stack->top].y1) / 2][(x
					+ stack->data[stack->top].x1) / 2];
					stack->data[stack->top].moved3 = boardp->map[y][x];

					DoMove(boardp, moves_leftp, stack->data[stack->top].x1, stack->data[stack->top].y1, x,
					y, leftp);
					/*do the move*/
					(*movesp)++; /*increment the moves counter*/
					sprintf(buffer, " Tiles Left: %d ", *leftp); /*update the tiles left
					display*/
					DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 0, buffer,
					A_REPLACE);
					sprintf(buffer, "  Moves: %d", *movesp); /*update the moves display*/
					DrawStr(calc(160, 240) - DrawStrWidth(buffer, F_4x6), 0, buffer,
					A_REPLACE);
					DrawPic(x, y, 9, 0, boardp->width, boardp->height); /*put the cursor
					back where it was*/
				}
				else
				{
					stack->top--;
				}
				blink = 0; /*stop the blinking*/
			}
			else /*if you are starting a move*/
			{
				if(boardp->map[y][x] != 0) /*if you did not press second on a blank
				space*/
				{
					dbg_printfln("writing start to top = %d", stack->top + 1);
					stack->data[++stack->top] = (MOVE){x, 0, y, 0, 0, 0, 0};
					if(FindLegalMoves(*boardp, legal_moves, x, y)) /*finds the legal
					moves for the square that the move was started on*/
					{
						blink = 1; /*turns on the blinking squares stuff if any legal
						moves were found*/
					}
				}
			}
			while(_rowread(0xFFFE) & calc(16, 1)); /*waits until second not pressed*/
		}
		if((key & calc(64, 2)) && !blink) /*if diamond pressed and no squares are
		blinking*/
		{
			if(UndoMove(stack, boardp, moves_leftp, x, y, movesp, leftp)) /*undo the last
			move on the stack, and if the cursor was erased*/
			{
				DrawPic(x, y, 9, 0, boardp->width, boardp->height); /*redraw the cursor*/
			}
			sprintf(buffer, " Tiles Left: %d ", *leftp); /*update the tiles left display*/
			DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 0, buffer,
			A_REPLACE);
			sprintf(buffer, "  Moves: %d", *movesp); /*updtae the moves left display*/
			DrawStr(calc(160, 240) - DrawStrWidth(buffer, F_4x6), 0, buffer, A_REPLACE);
			while(_rowread(0xFFFE) & calc(64, 2)); /*wait until diamond not pressed*/
		}
		if(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)) /*if esc pressed*/
		{
			go_on = 2; /*tell the loop to end*/
			*savep = RET_ESC;
			while(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)); /*wait until esc not
			pressed*/
		}
		if(blink && !Time) /*if the blink routine is active and it is time to blink*/
		{
			blinker ^= 1; /*switch wether the blinking squares are on or off*/
			Time += DELAY; /*restart the blink timer*/
			for(i = -1; i <= 1; i++)
			{
				for(j = -1; j <= 1; j++)
				{
					if(legal_moves[j + 1][i + 1])
					{
						DrawPic(stack->data[stack->top].x1 + 2 * i, stack->data[stack->top].y1 + 2 * j, 9,
						0, boardp->width, boardp->height); /*invert the blinking squares*/
					}
				}
			}
		}
		if((_rowread(calc(0xFFFD, 0xFF7F)) & calc(64, 32)) && !blink) /*if clear pressed
		and no squares are blinking*/
		{
			*savep = RET_SAVE;
			go_on = 2; /*tell the loop to quit*/
			while(_rowread(calc(0xFFFD, 0xFF7F)) & calc(64, 32)); /*wait until clear not
			pressed*/
		}
		if(!MoveLeft(*moves_leftp)) /*if there are no legal moves left*/
		{
			if(*leftp == 1) /*if there is also only one tile left*/
			{
				FontSetSys(F_8x10);
				DrawStr(calc(48, 88), calc(45, 59), "YOU WON!", A_REPLACE);
				FontSetSys(F_4x6);
				DrawStr(calc(44, 84), calc(94, 122), "Press 2nd to continue", A_NORMAL);
				Pause();
				go_on = 1; /*tell the loop to quit*/
			}
			else /*if you didn't achieve the perfect ending*/
			{
				go_on = GameEndMenu(*typep, lcd);
				if(go_on == 0)
				{
					if(UndoMove(stack, boardp, moves_leftp, x, y, movesp, leftp))
					{
						DrawPic(x, y, 9, 0, boardp->width, boardp->height);
					}
					sprintf(buffer, " Tiles Left: %d ", *leftp);
					DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 0, buffer,
					A_REPLACE);
					sprintf(buffer, "  Moves: %d", *movesp);
					DrawStr(calc(160, 240) - DrawStrWidth(buffer, F_4x6), 0, buffer,
					A_REPLACE);
				}
			}
		}
	} while(!go_on); /*loop until told not to*/

	stack->top = -1;

	return(go_on - 1); /*inform main of how many tiles were left and how many moves you took*/
}

void Shuffle(char array[], short length) /*saw this routine on the TICT message board.  it
randomly shuffles an array.  the arguments are the arrya and the length of the array*/
{
	short i;
	short j;

	for(i = 0; i < length; i++) /*iterate throught the whole list*/
	{
		j = random_long(length); /*randomly pick another element in the list*/
		array[i] ^= array[j] ^= array[i] ^= array[j]; /*swap the two elements.  this is
		another nifty trick I saw on the TICT messageboard*/
	}
}

void DoMove(BOARD *boardp, BOARD *moves_leftp, short x1, short y1, short x2, short y2,
short *leftp) /*the arguments are a pointer to the board that is being played on, a
pointer to the board that keeps track of legal moves, the coordinates of the beginning and
end of the move, and a pointer to the numer of tiles left.  many thanks to Marvin Thorman
for providing the logic in this function*/
{
	short temp; /*a variable that is used for calculating the correct results of the
	move*/

	temp = boardp->map[y1][x1] ^ boardp->map[(y1 + y2) / 2][(x1 + x2) / 2]; /*xor together
	the beginning of the move and the middle of the move*/
	if(temp > boardp->map[(y1 + y2) / 2][(x1 + x2) / 2]) /*if they were different
	primaries*/
	{
		temp = 0; /*the middle square is blank*/
	}
	(*leftp) -= (NumTiles(boardp->map[y1][x1]) + NumTiles(boardp->map[(y1 + y2) / 2][(x1 +
	x2) / 2]) + NumTiles(boardp->map[y2][x2])); /*adjust the number of tiles left*/
	boardp->map[y2][x2] |= boardp->map[y1][x1]; /*the end is ored with the beginning*/
	boardp->map[y1][x1] = 0; /*the beginning always becomes blank*/
	boardp->map[(y1 + y2) / 2][(x1 + x2) / 2] = temp; /*the middle is assigned the temp
	variable*/
	(*leftp) += (NumTiles(boardp->map[y1][x1]) + NumTiles(boardp->map[(y1 + y2) / 2][(x1 +
	x2) / 2]) + NumTiles(boardp->map[y2][x2])); /*update the number of tiles left*/
	DrawPic(x1, y1, boardp->map[y1][x1], 0, boardp->width, boardp->height); /*draw the new
	tile at the beginning of the move*/
	DrawPic((x1 + x2) / 2, (y1 + y2) / 2, boardp->map[(y1 + y2) / 2][(x1 + x2) / 2], 0,
	boardp->width, boardp->height); /*draw the new tile in the middle of the move*/
	DrawPic(x2, y2, boardp->map[y2][x2], 0, boardp->width, boardp->height); /*draw the new
	tile at the end of the move*/
	DrawPic(x2, y2, 10, 0, boardp->width, boardp->height); /*erase the border that the
	drawpic routine misses when drawing a tile*/
	UpdateLegalMoves(*boardp, moves_leftp); /*umm, you figure out what it does...*/
}

short FindLegalMoves(BOARD board, char legal_moves[][3], short x, short y) /*the arguments
are the board that is currently being played on , an array to pass back the legal moves,
and the coordinates of the beginning of the move.  returns the number of legal moves
found.  credit goes again to Marvin Thorman for the logic in this routine*/
{
	short i;
	short j;
	short temp = 0; /*the number of legal moves found*/

	for(i = -1; i <= 1; i++)
	{
		for(j = -1; j <= 1; j++)
		{
			if(x + 2 * i >= 0 && x + 2 * i < board.width && y + 2 * j >= 0 && y + 2 * j <
			board.height && board.map[y + j][x + i] != -1 && board.map[y + 2 * j][x + 2 *
			i] != -1 && (i || j)) /*if the current coordinates are within the shape and
			dimensions of the board and the offsets are not both zero*/
			{
				legal_moves[j + 1][i + 1] = 0; /*make sure that the relevant array element
				is set to zero*/
				switch(board.map[y][x])
				{
					case 0: /*no legal moves for a blank space*/
					break;

					case 1:
					case 2:
					case 4: /*primaries*/
					if(board.map[y + j][x + i] == 1 || board.map[y + j][x + i] == 2 ||
					board.map[y + j][x + i] == 4 || (board.map[y][x] & board.map[y + j][x
					+ i]) > 0) /*if the middle of the move is also a primary or has an
					element of the primary that is doing the jumping*/
					{
						if(board.map[y + 2 * j][x + 2 * i] == 0 || board.map[y + 2 * j][x
						+ 2 * i] == 1 || board.map[y + 2 * j][x + 2 * i] == 2 ||
						board.map[y + 2 * j][x + 2 * i] == 4 || (board.map[y][x] &
						board.map[y + 2 * j][x + 2 * i]) == 0) /*if the end of the move is
						a blank space, a primary, or does not have an element of the
						primary that is doing the jumping*/
						{
							legal_moves[j + 1][i + 1] = 1;
							/*found a legal move*/
							temp++;
						}
					}
					break;

					case 3:
					case 5:
					case 6: /*secondaries*/
					if(board.map[y + j][x + i] == board.map[y][x] || board.map[y + j][x +
					i] == 7) /*if the middle of the move is the same secodnary or the
					tertiary*/
					{
						if(board.map[y + 2 * j][x + 2 * i] == 0 || board.map[y + 2 * j][x
						+ 2 * i] == board.map[y][x] || (board.map[y][x] & board.map[y + 2
						* j][x + 2 * i]) == 0) /*if the end of the move is a blank space,
						the same secodnary, or a primary not contained in the secondary*/
						{
							legal_moves[j + 1][i + 1] = 1;
							/*found a legal move*/
							temp++;
						}
					}
					break;

					case 7: /*tertiary*/
					if(board.map[y + j][x + i] == 7) /*if the middle of the move is also a
					tertiary*/
					{
						if(board.map[y + 2 * j][x + 2 * i] == 0 || board.map[y + 2 * j][x
						+ 2 * i] == 7) /*if the end of the move is a blank space or a
						tertiary*/
						{
							legal_moves[j + 1][i + 1] = 1;
							/*found a legal move*/
							temp++;
						}
					}
				}
			}
			else
			{
				legal_moves[j + 1][i + 1] = 0; /*no legal moves here*/
			}
		}
	}
	return(temp);
}

char UndoMove(STACK *stack, BOARD *boardp, BOARD *moves_leftp, short x, short y, short
*movesp, short *leftp) /*the argumanets are the stack of moves to undo, a pointer to the
board that is cureently being played on, a pointer to the board that stores the number of
legal moves left, the coordinates of the cursor, and pointers to the number of moves left
and the number of tiles left*/
{
	char temp = 0; /*true if the cursor was erased*/

	dbg_printfln("undoing from top = %d", stack->top);
	if(stack->top != -1) /*if there are entries on the stack*/
	{
		(*movesp)--; /*decrement the number of moves taken*/
		(*leftp) -= (NumTiles(boardp->map[stack->data[stack->top].y1][stack->data[stack->top].x1]) +
		NumTiles(boardp->map[(stack->data[stack->top].y1 + stack->data[stack->top].y2) /
		2][stack->data[stack->top].x1 + stack->data[stack->top].x2 / 2]) +
		NumTiles(boardp->map[stack->data[stack->top].y2][stack->data[stack->top].x2]));
		/*adjust the number of tiles left*/

		boardp->map[stack->data[stack->top].y1][stack->data[stack->top].x1] =
		stack->data[stack->top].moved1; /*set the beginning of the move to what it was*/

		boardp->map[(stack->data[stack->top].y1 + stack->data[stack->top].y2) /
		2][(stack->data[stack->top].x1 + stack->data[stack->top].x2) / 2] =
		stack->data[stack->top].moved2; /*set the middle of the move to what it was*/

		boardp->map[stack->data[stack->top].y2][stack->data[stack->top].x2] =
		stack->data[stack->top].moved3; /*set the end of the move to what it was*/

		(*leftp) += (NumTiles(boardp->map[stack->data[stack->top].y1][stack->data[stack->top].x1]) +
		NumTiles(boardp->map[(stack->data[stack->top].y1 + stack->data[stack->top].y2) /
		2][(stack->data[stack->top].x1 + stack->data[stack->top].x2) / 2]) +
		NumTiles(boardp->map[stack->data[stack->top].y2][stack->data[stack->top].x2]));
		/*adjust the number of tiles left*/

		DrawPic(stack->data[stack->top].x1, stack->data[stack->top].y1,
		boardp->map[stack->data[stack->top].y1][stack->data[stack->top].x1], 0, boardp->width,
		boardp->height); /*update the graphics at the beginning of the move*/

		if(x == stack->data[stack->top].x1 && y == stack->data[stack->top].y1) /*if the cursor was
		erased*/
		{
			DrawPic(stack->data[stack->top].x1, stack->data[stack->top].y1, 10, 0, boardp->width,
			boardp->height);
			/*finish up the erasing job*/
			temp = 1; /*and tell playgame to redraw it*/
		}
		DrawPic((stack->data[stack->top].x1 + stack->data[stack->top].x2) / 2, (stack->data[stack->top].y1
		+ stack->data[stack->top].y2) / 2, boardp->map[(stack->data[stack->top].y1 +
		stack->data[stack->top].y2) / 2][(stack->data[stack->top].x1 + stack->data[stack->top].x2) / 2],
		0, boardp->width, boardp->height); /*update the graphics for the middle of the move*/

		if(x == (stack->data[stack->top].x1 + stack->data[stack->top].x2) / 2 && y ==
		(stack->data[stack->top].y1 + stack->data[stack->top].y2) / 2) /*if the cursor was erased*/
		{
			DrawPic((stack->data[stack->top].x1 + stack->data[stack->top].x2) / 2,
			(stack->data[stack->top].y1 + stack->data[stack->top].y2) / 2, 10, 0, boardp->width,
			boardp->height); /*finish up the erasing job*/
			temp = 1; /*tell playgame to redraw the cursor*/
		}
		DrawPic(stack->data[stack->top].x2, stack->data[stack->top].y2,
		boardp->map[stack->data[stack->top].y2][stack->data[stack->top].x2], 0, boardp->width,
		boardp->height); /*update the graphics at the end of the move*/

		if(x == stack->data[stack->top].x2 && y == stack->data[stack->top].y2) /*if the cursor was
		erased*/
		{
			DrawPic(stack->data[stack->top].x2, stack->data[stack->top].y2, 10, 0, boardp->width,
			boardp->height); /*finish up the erasing job*/
			temp = 1; /*tell playgame to redraw the cursor*/
		}
		stack->top--;
		UpdateLegalMoves(*boardp, moves_leftp);
	}
	dbg_printfln("after undo, top = %d", stack->top);
	return(temp);
}

void UpdateLegalMoves(BOARD board, BOARD *moves_leftp) /*updates the number of legal moves
available for each square.  arguments are the board currently being played on and a
pointer to the board that holds the legal move information*/
{
	char legal_moves[3][3]; /*just a dummy variable to satisfy the parameters of
	FindLegalMoves*/
	short i;
	short j;

	for(j = 0; j < board.height; j++)
	{
		for(i = 0; i < board.width; i++)
		{
			moves_leftp->map[j][i] = FindLegalMoves(board, legal_moves, i, j); /*update
			the number of moves left for all squares*/
		}
	}
}

char NumTiles(char tiles) /*returns the number of bits that are set, out of the first
three*/
{
	return((tiles & 1) + ((tiles & 2) >> 1) + ((tiles & 4) >> 2));
}

char MoveLeft(BOARD board) /*searches through the whole board, attempting to find a square
that has legal moves left.  returns true if it finds any.*/
{
	char found = 0; /*wether any legal moves have been found*/
	short i;
	short j;

	for(i = 0; i < board.width; i++)
	{
		for(j = 0; j < board.height; j++)
		{
			found = board.map[j][i];
			if(found > 0) /*if a square with legal moves has been found, there is no need
			to check the rest of the squares*/
			{
				break;
			}
		}
		if(found > 0)
		{
			break;
		}
	}
	return(found);
}

char GameEndMenu(GAME type, char *lcd) /*a menu of choices that pops up when you have no legal moves
left*/
{
	static char arrow[5] = {0x20, 0x10, 0xF8, 0x10, 0x20}; /*an arrow sprite*/
	char row = 0;
	short key;

	FastCopyScreen_R(LCD_MEM, lcd);
	ClrScr();
	DrawStr(calc(30, 70), 0, "You have no legal moves left.", A_NORMAL);
	DrawStr(calc(39, 79), 6, "What do you want to do?", A_NORMAL);
	DrawStr(5, 18, "Undo a move", A_NORMAL);
	if(type.tournament && type.board_num < 8) /*if you are playing a tournament game and
	you are not on the last board*/
	{
		DrawStr(5, 24, "Next board", A_NORMAL);
	}
	else
	{
		DrawStr(5, 24, "End game", A_NORMAL);
	}
	Sprite8_OR_R(0, 18, 5, arrow, LCD_MEM); /*draw the initial position of the arrow*/

	do
	{
		key = _rowread(0xFFFE); /*get the arrow keys*/

		if(key & calc(5, 160)) /*if up or down pressed*/
		{
			Sprite8_XOR_R(0, 18 + 6 * row, 5, arrow, LCD_MEM); /*erase the arrow*/
			row ^= 1; /*switch the row*/
			Sprite8_OR_R(0, 18 + 6 * row, 5, arrow, LCD_MEM); /*put the arrow in its new
			position*/
			while((_rowread(0xFFFE) & calc(1, 32)) || (_rowread(0xFFFE) & calc(4, 128)));
			/*wait until both up and down are not pressed*/
		}
	} while(!(_rowread(0xFFFE) & calc(16, 1))); /*loop until 2nd pressed*/
	while(_rowread(0xFFFE) & calc(16, 1)); /*wait until 2nd not pressed*/

	FastCopyScreen_R(lcd, LCD_MEM);

	return(row); /*tell playgame what choice you made*/
}

void CheckHighScores(GAME type, SCORE scores[][3], SCORE score) /*check to see if you have
a high score.  arguments are the type of game you played, the array of scores, and the
score for the game you just played*/
{
	short i;
	short index = (type.tournament ? 27 + type.board_num : 9 * type.difficulty +
	type.board_num); /*the index of the set of high scores to check*/

	for(i = 0; i < 3; i++)
	{
		if(score.left < scores[index][i].left) /*if the tiles left were less than the
		current high score*/
		{
			FontSetSys(F_8x10);
			DrawStr(calc(40, 80), calc(40, 54), "YOU HAVE A", A_REPLACE);
			DrawStr(calc(40, 80), calc(50, 64), "BEST SCORE", A_REPLACE);
			FontSetSys(F_4x6);
			memmove(&scores[index][i + 1], &scores[index][i], 2 - i);
			memcpy(&scores[index][i], &score, sizeof(SCORE));
			Pause();
			break;
		}
		else
		{
			if(score.left == scores[index][i].left && score.moves <=
			scores[index][i].moves) /*if the tiles left are the same but the number of
			moves taken is less*/
			{
				FontSetSys(F_8x10);
				DrawStr(calc(40, 80), calc(40, 54), "YOU HAVE A", A_REPLACE);
				DrawStr(calc(40, 80), calc(50, 64), "BEST SCORE", A_REPLACE);
				FontSetSys(F_4x6);
				memmove(&scores[index][i + 1], &scores[index][i], 2 - i);
				memcpy(&scores[index][i], &score, sizeof(SCORE));
				Pause();
				break;
			}
		}
	}
}